/* 2014, Tomas Barton, xbarton5@fi.muni.cz */

#include "superpositions.h"
#include <stdio.h>
#include <math.h>
#include "function.h"

#ifndef M_PI
#define M_PI           3.14159265358979323846
#endif


//global value
superposition_t superpositionData;
#define GAUSIAN_KERNEL_WIDTH 11

//debug file
#define WRITE_BINARY_DEBUG_FILE 0
FILE *debugFile = NULL;




/**
 * @brief tom_sum Computes sum of vector data.
 * @param data
 * @param size
 * @return sum of size elements in data[]
 */
float tom_sum(float * data, uint size){
    double sum = 0;
    while(size--){
        sum += *data++;
    }
    return (float) sum;
}

/**
 * @brief tom_gaussian_kernel Creates gaussian kernel.
 *
 * The kernel has length of size.
 *
 * sigma*3 = kernel[0] == kernel[size-1].
 *
 * Kernel is normalized: sum(kernel[]) == 1.0
 *
 * @param kernel (output) Array of at least size floats.
 * @param size size of kernel array, size must be odd number greater than 2
 * @return 1 on success
 */
int tom_gaussian_kernel(float * kernel, uint size)
{
    uint i;
    float sigma, sum;

    if(size % 2 == 0){
        fprintf(stderr, "Error: tom_gaussian_kernel(): size must be odd number.\n");
        return 0;
    }

    if(size == 1){
        kernel[0] = 1.0;
        return 1;
    }

    sigma = (size-1) / 6.0;

    for (i = 0; i <= size/2; i++){
        kernel[size/2 - i] = kernel[size/2 + i] = (1.0 / (sqrt(2*M_PI)*sigma)) * exp(-1.0 * (i*i) / (2*sigma*sigma));
    }

    sum = tom_sum(kernel, size);

    for (i = 0; i < size; ++i) {
        kernel[i] /= sum;
    }

    return 1;
}

/**
 * @brief tom_convolve Computes convolution between data and kernel. Boundaries are saturated to their
 * marginal values. (data[-1] == data[-2] ... == data[0])
 *
 * Kernel size must be an odd number because zero point of the kernel is in its center.
 *
 * @param data
 * @param size
 * @param output
 * @param kernel
 * @param kernel_size
 */
void tom_convolve(float *data, uint size, float * output, float * kernel, uint kernel_size){
    uint i;

    if(kernel_size % 2 == 0){
        fprintf(stderr, "Error: tom_iir_convolve() needs odd kernel_size.\n");
        return;
    }

    for (i = 0; i < size; ++i) {
        uint j;
        double sum = 0;
        int offset = i + kernel_size/2;

        for (j = 0; j < kernel_size; ++j) {
            if( (offset - (int)j) < 0 || (offset - (int)j) >= (int)size){
                // we are out of data boundaries -> lets use border values
                if ((offset - (int)j) < 0) sum += data[0]*kernel[j];
                else  sum += data[size-1]*kernel[j];
            } else {
                sum += data[offset-j]*kernel[j];
            }
        }
        output[i]= (float) sum;
    }
}

/**
 * @brief tom_diff Computes numerical derivation.
 * @param input
 * @param output array of the same size as input
 * @param size
 */
void tom_diff(float * input, float * output, uint size, uint step){
    uint i;

    for(i = 0; i < step; i++){
        output[i]=0.0;
    }

    for (i = step; i < size; ++i) {
        output[i] = input[i] - input[i-step];
    }
}

/**
 * @brief tom_superposition_init Computes Gaussian filter coefficients.
 * @return 0 on success
 */
int tom_superposition_init()
{
    int retval = tom_gaussian_kernel(superpositionData.filter_kernel, GAUSIAN_KERNEL_WIDTH);
    superpositionData.filter_size = GAUSIAN_KERNEL_WIDTH;

	//debug binary file open
	if (WRITE_BINARY_DEBUG_FILE)
	{
		debugFile = fopen("debugSuperpositionsFile.bin", "wb");
	}

    return retval;
}


typedef enum {
    BEFORE_TRIGGER, POSITIVE, NEGATIVE
} superposition_states_t;



#define SUPERPOSITION_MULT_CONST 20

/**
 * @brief tom_superposition_detection Computes superposition ratio
 * @param data Pulse data
 * @param size
 * @param super structure with configuration
 * @param amplitude amplitude of this signal
 * @return ration between amount of growth after maximum and amplitude
 */
int tom_superposition_detection(float *data, uint size, superposition_t *super)
{
    superposition_states_t state = BEFORE_TRIGGER;
    uint i;
    double sum= 0;

    /* Filter signal and compute numerical derivation */
    tom_convolve(data, size, super->filtered_buffer, super->filter_kernel, super->filter_size);
    tom_diff(super->filtered_buffer, super->diff_buffer, size, 1);

    for(i=0; i < size; i++){

        switch(state){
        case BEFORE_TRIGGER:
            if(data[i] > super->pulse_start_threshold){
                state = POSITIVE;
            }
            break;

        case POSITIVE:
            if(super->diff_buffer[i] < -1.0){
                state = NEGATIVE;
            }
            break;

        case NEGATIVE:

            /* In NEGATIVE state the pulse should be after its maximum and therefore descending
             * which means that the derivative should be negative. Positive derivative can be caused
             * by noise or superposition of another pulse. By filtering we minimize noise influence
             * and in *ideal* case only superpositions are left. The sum variable will express how
             * much has the pulse risen after its maximum.
             */

            if(super->diff_buffer[i] > 0.0){
                sum += super->diff_buffer[i];
            }

            break;
        }
    }

    /* The ratio variable expresses the ratio between amplitude of the pulse and the amount of growth of
     * the signal in a region after its maximum. In case of well filtered signal the ratio above 0.05
     * indicates presence of aditional superpositioned pulses.
     */
	 return sum * SUPERPOSITION_MULT_CONST;
}

//vrati, zda se jedna o superponovany pulz
int DetekujSuperpozice(float* impuls, int delka, int thresholdBodEnergie0, int thresholdBodEnergieMax, float amplitudaPulzu, float discrParametr, int *parametrSuperpozice)
{
	*parametrSuperpozice = tom_superposition_detection(impuls, delka, &superpositionData);

	//zapis binarniho souboru pro debug superpozic
	if (WRITE_BINARY_DEBUG_FILE)
	{
		unsigned short tomPulse[3];
		tomPulse[0] = (unsigned short)amplitudaPulzu;
		tomPulse[1] = (unsigned short)discrParametr;
		tomPulse[2] = (unsigned short)*parametrSuperpozice;
		fwrite(tomPulse, sizeof(unsigned short), 3, debugFile);
	}

	//superpozice se oddluj pmkou, offset je thresholdBodEnergie0, smrnice thresholdBodEnergieMax / MATRIX_SIZE
	return (*parametrSuperpozice > (thresholdBodEnergie0 + amplitudaPulzu*(thresholdBodEnergieMax / (float)MATRIX_SIZE))) ? 1 : 0;
}




